function candlesSTD(candles, field = 'close') {
    if (!candles || candles.length === 0) {
        return 0; // Return 0 for an empty set or invalid input
    }

    let mean = 0;
    for (let i = 0; i < candles.length; i++) {
        mean += candles[i][field];
    }
    mean /= candles.length;

    let variance = 0;
    for (let i = 0; i < candles.length; i++) {
        const diffFromMean = candles[i][field] - mean;
        variance += diffFromMean * diffFromMean;
    }
    variance /= candles.length;

    const standardDeviation = Math.sqrt(variance);
    return standardDeviation;
}


// Example usage
// const candles = [
//     { open: 100, high: 110, low: 90, close: 105, volume: 1000 },
//     // ... more candles ...
// ];
//
// const fieldToCalculate = 'close'; // You can change this to the desired field
//
// const std = calculateStandardDeviation(candles, fieldToCalculate);
// console.log(`Standard Deviation of ${fieldToCalculate}: ${std}`);

function calculateVWAP(candles, doSMA=false) {
    let totalVolume = 0;
    let vwapNumerator = 0;

    for (let i = 0; i < candles.length; i++) {
        const candle = candles[i];
        const price = (candle.high + candle.low + candle.close) / 3;  // average price of the candle
        const volume = doSMA ? 1 : candle.volume;

        totalVolume += volume;
        vwapNumerator += price * volume;
    }

    if (totalVolume === 0) {
        // Avoid division by zero. This could happen if all candles have zero volume.
        return 0;
    }

    return vwapNumerator / totalVolume;
}

function calculateRR(candles, periods=15, rrLimit = 50, includeRecentNCandles = false) {
    const resultCandles = [];

    for (let i = 0; i < candles.length; i++) {
        const currentCandle = candles[i];
        const futureCandles = candles.slice(i + 1, i + 1 + periods);
        const pastCandles = candles.slice(Math.max(0, i - periods + 1), i+1); // Select past candles [including current one, remove last +1 to exclude]

        var futureSMA = calculateVWAP(futureCandles, true);
        var futureVWAP = calculateVWAP(futureCandles, false);
        var futureSMAFraction = futureSMA/currentCandle.close;
        var futureVWAPFraction = futureVWAP/currentCandle.close;
        var futureSMAFractionLog2 = Math.log2(futureSMA/currentCandle.close);
        var futureVWAPFractionLog2 = Math.log2(futureVWAP/currentCandle.close);

        var std = candlesSTD(pastCandles);

        var futureSMA_STD = (futureSMA-currentCandle.close)/(std || 1);
        var futureVWAP_STD = (futureVWAP-currentCandle.close)/(std || 1);

        if (futureCandles.length < periods) {
            // Not enough future candles for calculation
            resultCandles.push({ ...currentCandle });
            continue;
        }

        const futureHighs = futureCandles.map(candle => candle.high);
        const futureLows = futureCandles.map(candle => candle.low);
        let futureHigh = -Infinity;
        let futureLow = Infinity;
        let futureHighIndex = -1;
        let futureLowIndex = -1;

        for (let i = 0; i < futureHighs.length; i++) {
            if (futureHighs[i] > futureHigh) {
                futureHigh = futureHighs[i];
                futureHighIndex = i+1;
            }
            if (futureLows[i] < futureLow) {
                futureLow = futureLows[i];
                futureLowIndex = i+1;
            }
        }

        const closePrice = currentCandle.close;
        const lossGain = [futureLow - closePrice, futureHigh - closePrice];
        const lossGainFraction = [futureLow/closePrice,futureHigh/closePrice];
        const lossGainFractionLog2 = lossGainFraction.map(Math.log2);

        const lossGainSTD = [lossGain[0]/std, lossGain[1]/std];
        const lossGainNormalized = [lossGain[0]/closePrice, lossGain[1]/closePrice]; //note that this is slightly diff than downUpDrawMaxPercent -- $100 -> [95,105] becomes [0.05,0.05] rather than [0.95, 1.05]

        const riskReward = Math.min(rrLimit, lossGain[1] / Math.max(1.0/rrLimit,-lossGain[0])); //"positive" losses are counted as zero , basically should return infinite RR in that situation

        const updatedCandle = {
            ...currentCandle
        };

        updatedCandle[`rr_${periods}`] = {
            futureSMA,futureSMAFraction, futureSMAFractionLog2, futureSMA_STD,
            futureVWAP, futureVWAPFraction, futureVWAPFractionLog2,futureVWAP_STD ,
            lossGainAbsolute: lossGain,
            lossGainSTD,
            lossGainNormalized,
            lossGainFraction,
            lossGainFractionLog2,
            riskReward,
            riskRewardLog2: Math.max(-1000,Math.min(1000,Math.log2(riskReward))) ,
            futureHighIndex,
            futureLowIndex,
            stdPast: std
        }

        resultCandles.push(updatedCandle);
    }

    if (includeRecentNCandles) {
        for (let i = 0; i < periods - 1; i++) {
            const recentCandle = candles[i];
            const updatedCandle = {
                ...recentCandle
            };

            updatedCandle[`rr_${periods}`] = {
                futureSMA,
                futureVWAP,
                lossGainAbsolute: [0,0],
                lossGainSTD: [0,0],
                lossGainNormalized: [1,1],
                lossGainFraction: [1,1],
                lossGainFractionLog2: [0,0],
                riskReward: 1,
                riskRewardLog2: 0 ,
                futureHighIndex:0,
                futureLowIndex:0
            }
            resultCandles[i] = updatedCandle;
        }
    }

    return resultCandles;
}

module.exports = {calculateRR}

// // Example usage
// const candles = [
//     { open: 100, high: 110, low: 90, close: 105, volume: 1000 },
//     // ... more candles ...
// ];
//
// const periods = 15;
// const includeRecentNCandles = true;
// const candlesWithDraws = calculateDraws(candles, periods, includeRecentNCandles);
//
// console.log(candlesWithDraws);
